"""
QS Support.


Adding support for Querysource related functions as datasource support to components.
"""
import os
from abc import ABC
from collections.abc import Callable
import asyncio
from importlib import import_module
from navconfig import config
from sqlalchemy.orm import Session
from sqlalchemy import create_engine, text
from sqlalchemy.pool import NullPool
from asyncdb import AsyncDB
from asyncdb.drivers.pg import pg
from querysource.datasources.drivers import DataDriver
from querysource.conf import (
    default_dsn,
    sqlalchemy_url,
    DB_TIMEOUT,
    DB_STATEMENT_TIMEOUT,
    DB_SESSION_TIMEOUT,
    DB_KEEPALIVE_IDLE,
    POSTGRES_TIMEOUT
)
from ...exceptions import ConfigError, ComponentError

class QSSupport(ABC):
    """QSSupport.

    Adding Support for Querysource parameters.
    """
    use_sqlalchemy: bool = False

    def get_config_value(self, key_name, default: str = None):
        if key_name is None:
            return default
        if val := os.getenv(str(key_name)):
            return val
        if val := config.get(str(key_name), default):
            return val
        else:
            # TODO: get from replacing masks or memecached
            return key_name

    def processing_credentials(self):
        if self.credentials:
            for key, value in self.credentials.items():
                default = getattr(self, key, value)
                try:
                    val = self.get_config_value(
                        key_name=value, default=default
                    )
                    self.credentials[key] = val
                except (TypeError, KeyError) as err:
                    raise ConfigError(
                        f"{__name__}: Wrong or missing Credentials"
                    ) from err
        return self.credentials

    def get_connection(
        self,
        driver: str = 'pg',
        dsn: str = None,
        params: dict = None,
        **kwargs
    ) -> Callable:
        """Useful for internal connections of QS.
        """
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.get_running_loop()
        if not dsn:
            dsn = default_dsn
        if driver == 'pg':
            args: dict = {
                "min_size": 2,
                "server_settings": {
                    "application_name": "FlowTask:CopyTo",
                    "client_min_messages": "notice",
                    "jit": "off",
                    "statement_timeout": f"{DB_STATEMENT_TIMEOUT}",
                    "idle_session_timeout": f"{DB_SESSION_TIMEOUT}",
                    "effective_cache_size": "2147483647",
                    "tcp_keepalives_idle": f"{DB_KEEPALIVE_IDLE}",
                },
                "timeout": int(POSTGRES_TIMEOUT),
                **kwargs
            }
        else:
            args = {
                "timeout": int(DB_TIMEOUT)
            }
        return AsyncDB(
            driver,
            dsn=dsn,
            params=params,
            loop=loop,
            **args
        )

    def default_connection(self, dsn: str = None):
        try:
            timeout = int(DB_TIMEOUT)
        except TypeError:
            timeout = 360
        if not dsn:
            dsn = default_dsn
        try:
            kwargs: dict = {
                "min_size": 2,
                "server_settings": {
                    "application_name": "FlowTask:CopyToPg",
                    "client_min_messages": "notice",
                    "jit": "off",
                    "statement_timeout": f"{DB_STATEMENT_TIMEOUT}",
                    "idle_session_timeout": f"{DB_SESSION_TIMEOUT}",
                    "effective_cache_size": "2147483647",
                    "tcp_keepalives_idle": f"{DB_KEEPALIVE_IDLE}",
                },
                "timeout": timeout,
            }
            self._connection = pg(dsn=dsn, loop=self._loop, **kwargs)
            return self._connection
        except Exception as err:
            raise ComponentError(
                f"Error configuring Pg Connection: {err!s}"
            ) from err

    def get_driver(self, driver) -> DataDriver:
        """Getting a Database Driver from Datasource Drivers.
        """
        # load dynamically
        clspath = f'querysource.datasources.drivers.{driver}'
        clsname = f'{driver}Driver'
        try:
            self._dsmodule = import_module(clspath)
            return getattr(self._dsmodule, clsname)
        except (AttributeError, ImportError) as ex:
            raise RuntimeError(
                f"QS: There is no Driver {driver}: {ex}"
            ) from ex

    async def get_datasource(self, name: str):
        """get_datasource.

        Get the datasource from the database.
        """
        try:
            db = self.default_connection()
            async with await db.connection() as conn:
                sql = f"SELECT * FROM public.datasources WHERE name = '{name}'"
                row, error = await conn.queryrow(sql)
                if error:
                    self._logger.warning(f'DS Error: {error}')
                    return False
                try:
                    driver = self.get_driver(row['driver'])
                    # TODO: encrypting credentials in database:
                    if row['dsn']:
                        data = {
                            "dsn": row['dsn']
                        }
                    else:
                        try:
                            data = {
                                **dict(row['params']),
                            }
                        except TypeError:
                            data = dict(row['params'])
                        for key, val in row.get('credentials', {}).items():
                            data[key] = self.get_config_value(
                                key_name=val,
                                default=val
                            )
                        for key, val in row.get('params', {}).items():
                            data[key] = self.get_config_value(
                                key_name=val,
                                default=val
                            )
                    return driver(**data)
                except Exception as ex:  # pylint: disable=W0703
                    self._logger.error(ex)
                    return False
        except Exception as exc:
            self._logger.error(exc)
            return False

    def get_sqlalchemy_connection(self, dsn: str = None):
        if not dsn:
            dsn = sqlalchemy_url
        self._engine = create_engine(dsn, echo=False, poolclass=NullPool)
        self._connection = Session(self._engine)
        return self._connection

    async def create_connection(self):
        if hasattr(self, "credentials"):
            credentials = self.processing_credentials()
            return self.get_connection(
                driver='pg',
                params=credentials
            )
        elif hasattr(self, 'dsn'):
            dsn = self.get_config_value(self.dsn, self.dsn)
            if self.use_sqlalchemy is True:
                if dsn.startswith('postgres:'):
                    dsn = dsn.replace('postgres:', 'postgresql:')
                    return self.get_sqlalchemy_connection(
                        dsn=dsn
                    )
            return self.get_connection(
                driver='pg',
                dsn=dsn
            )
        elif hasattr(self, "datasource"):
            datasource = await self.get_datasource(name=self.datasource)
            if datasource.driver_type == 'asyncdb':
                driver = datasource.driver
                return AsyncDB(
                    driver,
                    dsn=datasource.dsn,
                    params=datasource.params()
                )
            else:
                raise ConfigError(
                    f"Invalid Datasource type {datasource.driver_type} for {self.datasource}"
                )
        else:
            return self.default_connection()
